from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger, time_it

from .config_slurper import SPLKConfigData
from .kconfig import KConfig


@dataclass
class EditCommandConfig(DataClassDictMixin):
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={
            "help": "Project root directory. "
            "Defaults to the current directory if not specified."
        },
    )
    kconfig_model_file: Optional[Path] = field(
        default=None, metadata={"help": "KConfig model file (KConfig)."}
    )
    kconfig_config_file: Optional[Path] = field(
        default=None, metadata={"help": "KConfig user configuration file (config.txt)."}
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "EditCommandConfig":
        return cls.from_dict(vars(namespace))


class EditCommand(Command):
    def __init__(self) -> None:
        super().__init__("edit", "Edit KConfig configuration.")
        self.logger = logger.bind()

    @time_it("Build")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        cmd_config = EditCommandConfig.from_namespace(args)
        if cmd_config.kconfig_model_file is None:
            kconfig_data = SPLKConfigData(cmd_config.project_dir)
            variants = kconfig_data.get_variants()
            variant_names = [variant.name for variant in variants]
            selected_variant = self._select_variant(variant_names)
            if selected_variant is not None:
                variant_data = kconfig_data.find_variant_config(selected_variant)
                if variant_data is not None:
                    variant_data.config.menu_config()
                else:
                    self.logger.error(f"Variant {selected_variant} not found.")
        else:
            KConfig(
                cmd_config.kconfig_model_file, cmd_config.kconfig_config_file
            ).menu_config()
        return 0

    def _select_variant(self, variant_names: List[str]) -> Optional[str]:
        """Print the list of variants to choose from and let the user choose one"""
        selected_variant = None
        self.logger.info("Select a variant:")
        for index, variant_name in enumerate(variant_names):
            self.logger.info(f" [{index + 1}] {variant_name}")
        while True:
            try:
                selected_variant = variant_names[
                    int(input(f"Select a variant (1-{len(variant_names)}): ")) - 1
                ]
                break
            except (ValueError, IndexError):
                self.logger.error(
                    "Invalid input. Please try again or press CTRL+C to abort."
                )
            # In case of KeyboardInterrupt (CTRL+C) we want to abort
            except KeyboardInterrupt:
                self.logger.warning("Aborted by user.")
                break
        return selected_variant

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, EditCommandConfig)



import sys
from argparse import ArgumentParser
from sys import argv

from py_app_dev.core.cmd_line import CommandLineHandlerBuilder
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger, setup_logger

from kspl import __version__
from kspl.edit import EditCommand
from kspl.generate import GenerateCommand
from kspl.gui import GuiCommand


def do_run() -> None:
    parser = ArgumentParser(
        prog="kspl", description="kconfig for SPL", exit_on_error=False
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"%(prog)s {__version__}"
    )
    builder = CommandLineHandlerBuilder(parser)
    builder.add_commands([GuiCommand(), GenerateCommand(), EditCommand()])
    handler = builder.create()
    handler.run(argv[1:])


def main() -> int:
    try:
        setup_logger()
        do_run()
    except UserNotificationException as e:
        logger.error(f"{e}")
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



import tkinter
from abc import abstractmethod
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from enum import auto
from pathlib import Path
from tkinter import simpledialog, ttk
from typing import Any, Dict, List, Optional

import customtkinter
from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger, time_it
from py_app_dev.mvp.event_manager import EventID, EventManager
from py_app_dev.mvp.presenter import Presenter
from py_app_dev.mvp.view import View

from kspl.config_slurper import SPLKConfigData, VariantViewData
from kspl.kconfig import ConfigElementType, EditableConfigElement, TriState


class KSplEvents(EventID):
    EDIT = auto()


class CTkView(View):
    @abstractmethod
    def mainloop(self) -> None:
        pass


@dataclass
class EditEventData:
    variant: VariantViewData
    config_element_name: str
    new_value: Any


class MainView(CTkView):
    def __init__(
        self,
        event_manager: EventManager,
        elements: List[EditableConfigElement],
        variants: List[VariantViewData],
    ) -> None:
        self.event_manager = event_manager
        self.elements = elements
        self.elements_dict = {elem.name: elem for elem in elements}
        self.variants = variants

        self.logger = logger.bind()
        self.edit_event_data: Optional[EditEventData] = None
        self.trigger_edit_event = self.event_manager.create_event_trigger(
            KSplEvents.EDIT
        )
        self.root = customtkinter.CTk()

        # Configure the main window
        self.root.title("K-SPL")
        self.root.geometry(f"{1080}x{580}")

        # ========================================================
        # create tabview and populate with frames
        tabview = customtkinter.CTkTabview(self.root)
        self.tree = self.create_tree_view(tabview.add("Configuration"))
        self.tree["columns"] = tuple(variant.name for variant in self.variants)
        self.tree.heading("#0", text="Configuration")
        for variant in self.variants:
            self.tree.heading(variant.name, text=variant.name)
        # Keep track of the mapping between the tree view items and the config elements
        self.tree_view_items_mapping = self.populate_tree_view()
        self.tree.pack(fill="both", expand=True)
        # TODO: make the tree view editable
        # self.tree.bind("<Double-1>", self.double_click_handler)

        # ========================================================
        # put all together
        self.root.grid_columnconfigure(0, weight=1)
        self.root.grid_rowconfigure(0, weight=1)
        tabview.grid(row=0, column=0, sticky="nsew")

    def mainloop(self) -> None:
        self.root.mainloop()

    def create_tree_view(self, frame: customtkinter.CTkFrame) -> ttk.Treeview:
        frame.grid_rowconfigure(0, weight=10)
        frame.grid_rowconfigure(1, weight=1)
        frame.grid_columnconfigure(0, weight=1)

        columns = [var.name for var in self.variants]

        style = ttk.Style()
        style.configure(
            "mystyle.Treeview",
            highlightthickness=0,
            bd=0,
            font=("Calibri", 14),
            rowheight=30,
        )  # Modify the font of the body
        style.configure(
            "mystyle.Treeview.Heading", font=("Calibri", 14, "bold")
        )  # Modify the font of the headings

        # create a Treeview widget
        config_treeview = ttk.Treeview(
            frame,
            columns=columns,
            show="tree headings",
            style="mystyle.Treeview",
        )
        config_treeview.grid(row=0, column=0, sticky="nsew")
        return config_treeview

    def populate_tree_view(self) -> Dict[str, str]:
        """
        Populates the tree view with the configuration elements.
        :return: a mapping between the tree view items and the configuration elements
        """
        stack = []  # To keep track of the parent items
        last_level = -1
        mapping: Dict[str, str] = {}

        for element in self.elements:
            values = self.collect_values_for_element(element)
            if element.level == 0:
                # Insert at the root level
                item_id = self.tree.insert("", "end", text=element.name, values=values)
                stack = [item_id]  # Reset the stack with the root item
            elif element.level > last_level:
                # Insert as a child of the last inserted item
                item_id = self.tree.insert(
                    stack[-1], "end", text=element.name, values=values
                )
                stack.append(item_id)
            elif element.level == last_level:
                # Insert at the same level as the last item
                item_id = self.tree.insert(
                    stack[-2], "end", text=element.name, values=values
                )
                stack[-1] = item_id  # Replace the top item in the stack
            else:
                # Go up in the hierarchy and insert at the appropriate level
                item_id = self.tree.insert(
                    stack[element.level - 1], "end", text=element.name, values=values
                )
                stack = stack[: element.level] + [item_id]

            last_level = element.level
            mapping[item_id] = element.name
        return mapping

    def collect_values_for_element(
        self, element: EditableConfigElement
    ) -> List[int | str]:
        return (
            [
                self.prepare_value_to_be_displayed(
                    element.type, variant.config_dict.get(element.name, None)
                )
                for variant in self.variants
            ]
            if not element.is_menu
            else []
        )

    def prepare_value_to_be_displayed(
        self, element_type: ConfigElementType, value: Any
    ) -> str:
        """
        UNKNOWN  - N/A
        BOOL     - ✅ ⛔
        TRISTATE - str
        STRING   - str
        INT      - str
        HEX      - str
        MENU     - N/A
        """
        if value is None:
            return "N/A"
        elif element_type == ConfigElementType.BOOL:
            return "✅" if value == TriState.Y else "⛔"
        else:
            return str(value)

    def double_click_handler(self, event: tkinter.Event) -> None:  # type: ignore
        current_selection = self.tree.selection()
        if not current_selection:
            return

        selected_item = current_selection[0]
        selected_element_name = self.tree_view_items_mapping[selected_item]

        variant_idx_str = self.tree.identify_column(event.x)  # Get the clicked column
        variant_idx = (
            int(variant_idx_str.split("#")[-1]) - 1
        )  # Convert to 0-based index

        if variant_idx < 0 or variant_idx >= len(self.variants):
            return

        selected_variant = self.variants[variant_idx]
        selected_element = self.elements_dict[selected_element_name]
        selected_element_value = selected_variant.config_dict.get(selected_element_name)

        # TODO: Consider the actual configuration type (ConfigElementType)
        if not selected_element.is_menu:
            new_value: Any = None
            if selected_element.type == ConfigElementType.BOOL:
                # Toggle the boolean value
                new_value = (
                    TriState.N if selected_element_value == TriState.Y else TriState.Y
                )
            elif selected_element.type == ConfigElementType.INT:
                tmp_int_value = simpledialog.askinteger(
                    "Enter new value",
                    "Enter new value",
                    initialvalue=selected_element_value,
                )
                if tmp_int_value is not None:
                    new_value = tmp_int_value
            else:
                # Prompt the user to enter a new string value using messagebox
                tmp_str_value = simpledialog.askstring(
                    "Enter new value",
                    "Enter new value",
                    initialvalue=str(selected_element_value),
                )
                if tmp_str_value is not None:
                    new_value = tmp_str_value

            # Check if the value has changed
            if new_value:
                # Trigger the EDIT event
                self.create_edit_event_trigger(
                    selected_variant, selected_element_name, new_value
                )

    def create_edit_event_trigger(
        self, variant: VariantViewData, element_name: str, new_value: Any
    ) -> None:
        self.edit_event_data = EditEventData(variant, element_name, new_value)
        self.trigger_edit_event()

    def pop_edit_event_data(self) -> Optional[EditEventData]:
        result = self.edit_event_data
        self.edit_event_data = None
        return result


class KSPL(Presenter):
    def __init__(self, event_manager: EventManager, project_dir: Path) -> None:
        self.event_manager = event_manager
        self.event_manager.subscribe(KSplEvents.EDIT, self.edit)
        self.logger = logger.bind()
        self.kconfig_data = SPLKConfigData(project_dir)
        self.view = MainView(
            self.event_manager,
            self.kconfig_data.get_elements(),
            self.kconfig_data.get_variants(),
        )

    def edit(self) -> None:
        edit_event_data = self.view.pop_edit_event_data()
        if edit_event_data is None:
            self.logger.error("Edit event received but event data is missing!")
        else:
            self.logger.debug(
                "Edit event received: "
                f"'{edit_event_data.variant.name}:{edit_event_data.config_element_name} = {edit_event_data.new_value}'"
            )
            # Update the variant configuration data with the new value
            variant = self.kconfig_data.find_variant_config(
                edit_event_data.variant.name
            )
            if variant is None:
                raise ValueError(
                    f"Could not find variant '{edit_event_data.variant.name}'"
                )
            config_element = variant.find_element(edit_event_data.config_element_name)
            if config_element is None:
                raise ValueError(
                    f"Could not find config element '{edit_event_data.config_element_name}'"
                )
            config_element.value = edit_event_data.new_value

    def run(self) -> None:
        self.view.mainloop()


@dataclass
class GuiCommandConfig(DataClassDictMixin):
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={
            "help": "Project root directory. "
            "Defaults to the current directory if not specified."
        },
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "GuiCommandConfig":
        return cls.from_dict(vars(namespace))


class GuiCommand(Command):
    def __init__(self) -> None:
        super().__init__("view", "View all SPL KConfig configurations.")
        self.logger = logger.bind()

    @time_it("Build")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        config = GuiCommandConfig.from_namespace(args)
        event_manager = EventManager()
        KSPL(event_manager, config.project_dir.absolute()).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, GuiCommandConfig)



"""Used to run main from the command line when run from this repository.
This is required because kspl module is not visible when running
from the repository."""
import runpy
import sys
from pathlib import Path

sys.path.insert(0, Path(__file__).parent.parent.absolute().as_posix())
runpy.run_module("kspl.main", run_name="__main__")



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger

from kspl.kconfig import EditableConfigElement, KConfig


@dataclass
class VariantViewData:
    """A variant is a set of configuration values for a KConfig model."""

    name: str
    config_dict: Dict[str, Any]


@dataclass
class VariantData:
    name: str
    config: KConfig

    def find_element(self, element_name: str) -> Optional[EditableConfigElement]:
        return self.config.find_element(element_name)


class SPLKConfigData:
    def __init__(self, project_root_dir: Path) -> None:
        self.project_root_dir = project_root_dir.absolute()
        variant_config_files = self._search_variant_config_file(self.project_root_dir)
        if not self.kconfig_model_file.is_file():
            raise UserNotificationException(
                f"File {self.kconfig_model_file} does not exist."
            )
        self.model = KConfig(self.kconfig_model_file)
        if variant_config_files:
            self.variant_configs: List[VariantData] = [
                VariantData(
                    self._get_variant_name(file), KConfig(self.kconfig_model_file, file)
                )
                for file in variant_config_files
            ]
        else:
            self.variant_configs = [VariantData("Default", self.model)]
        self.logger = logger.bind()

    @property
    def kconfig_model_file(self) -> Path:
        return self.project_root_dir / "KConfig"

    def get_elements(self) -> List[EditableConfigElement]:
        return self.model.elements

    def get_variants(self) -> List[VariantViewData]:
        variants = []

        for variant in self.variant_configs:
            variants.append(
                VariantViewData(
                    variant.name,
                    {
                        config_elem.name: config_elem.value
                        for config_elem in variant.config.elements
                        if not config_elem.is_menu
                    },
                )
            )
        return variants

    def _get_variant_name(self, file: Path) -> str:
        return file.relative_to(self.project_root_dir / "variants").parent.as_posix()

    def _search_variant_config_file(self, project_dir: Path) -> List[Path]:
        """
        Finds all files called 'config.txt' in the variants directory
        and returns a list with their paths.
        """
        return list((project_dir / "variants").glob("**/config.txt"))

    def find_variant_config(self, variant_name: str) -> Optional[VariantData]:
        for variant in self.variant_configs:
            if variant.name == variant_name:
                return variant
        return None



import json
from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

import kconfiglib
from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.logging import logger, time_it

from kspl.kconfig import ConfigElementType, ConfigurationData, KConfig, TriState


class GeneratedFile:
    def __init__(
        self, path: Path, content: str = "", skip_writing_if_unchanged: bool = False
    ) -> None:
        self.path = path

        self.content = content

        self.skip_writing_if_unchanged = skip_writing_if_unchanged

    def to_string(self) -> str:
        return self.content

    def to_file(self) -> None:
        """Only write to file if the content has changed.
        The directory of the file is created if it does not exist."""

        content = self.to_string()

        if (
            not self.path.exists()
            or not self.skip_writing_if_unchanged
            or self.path.read_text() != content
        ):
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self.path.write_text(content)


class FileWriter(ABC):
    """- writes the ConfigurationData to a file"""

    def __init__(self, output_file: Path):
        self.output_file = output_file

    def write(self, configuration_data: ConfigurationData) -> None:
        """- writes the ConfigurationData to a file
        The file shall not be modified if the content is the same as the existing one"""
        content = self.generate_content(configuration_data)
        GeneratedFile(
            self.output_file, content, skip_writing_if_unchanged=True
        ).to_file()

    @abstractmethod
    def generate_content(self, configuration_data: ConfigurationData) -> str:
        """- generates the content of the file from the ConfigurationData"""


class HeaderWriter(FileWriter):
    """Writes the ConfigurationData as pre-processor defines in a C Header file"""

    config_prefix = "CONFIG_"  # Prefix for all configuration defines

    def generate_content(self, configuration_data: ConfigurationData) -> str:
        """This method does exactly what the kconfiglib.write_autoconf() method does.
        We had to implemented here because we refactor the file writers to use the ConfigurationData
        instead of the KConfig configuration. ConfigurationData has variable substitution already done.
        """
        result: List[str] = [
            "/** @file */",
            "#ifndef __autoconf_h__",
            "#define __autoconf_h__",
            "",
        ]

        def add_define(define_decl: str, description: str) -> None:
            result.append(f"/** {description} */")
            result.append(define_decl)

        for element in configuration_data.elements:
            val = element.value
            if element.type in [ConfigElementType.BOOL, ConfigElementType.TRISTATE]:
                if val == TriState.Y:
                    add_define(
                        f"#define {self.config_prefix}{element.name} 1",
                        element.name,
                    )
                elif val == TriState.M:
                    add_define(
                        "#define {}{}_MODULE 1".format(
                            self.config_prefix, element.name
                        ),
                        element.name,
                    )

            elif element.type is ConfigElementType.STRING:
                add_define(
                    '#define {}{} "{}"'.format(
                        self.config_prefix, element.name, kconfiglib.escape(val)
                    ),
                    element.name,
                )

            else:  # element.type in [INT, HEX]:
                if element.type is ConfigElementType.HEX:
                    val = hex(val)
                add_define(
                    f"#define {self.config_prefix}{element.name} {val}",
                    element.name,
                )
        result.extend(["", "#endif /* __autoconf_h__ */", ""])
        return "\n".join(result)


class JsonWriter(FileWriter):
    """Writes the ConfigurationData in json format"""

    def generate_content(self, configuration_data: ConfigurationData) -> str:
        result = {}
        for element in configuration_data.elements:
            if element.type is ConfigElementType.BOOL:
                result[element.name] = True if element.value == TriState.Y else False
            else:
                result[element.name] = element.value
        return json.dumps({"features": result}, indent=4)


class CMakeWriter(FileWriter):
    """Writes the ConfigurationData as CMake variables"""

    def generate_content(self, configuration_data: ConfigurationData) -> str:
        result: List[str] = []
        add = result.append
        for element in configuration_data.elements:
            val = element.value
            if element.type is ConfigElementType.BOOL:
                val = True if element.value == TriState.Y else False
            add(f'set({element.name} "{val}")')

        return "\n".join(result)


@dataclass
class GenerateCommandConfig(DataClassDictMixin):
    kconfig_model_file: Path = field(metadata={"help": "KConfig model file (KConfig)."})
    kconfig_config_file: Optional[Path] = field(
        default=None, metadata={"help": "KConfig user configuration file (config.txt)."}
    )
    out_header_file: Optional[Path] = field(
        default=None, metadata={"help": "File to write the configuration as C header."}
    )
    out_json_file: Optional[Path] = field(
        default=None,
        metadata={"help": "File to write the configuration in JSON format."},
    )
    out_cmake_file: Optional[Path] = field(
        default=None,
        metadata={"help": "File to write the configuration in CMake format."},
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "GenerateCommandConfig":
        return cls.from_dict(vars(namespace))


class GenerateCommand(Command):
    def __init__(self) -> None:
        super().__init__(
            "generate", "Generate the KConfig configuration in the specified formats."
        )
        self.logger = logger.bind()

    @time_it("Build")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        cmd_config = GenerateCommandConfig.from_namespace(args)
        config = KConfig(
            cmd_config.kconfig_model_file, cmd_config.kconfig_config_file
        ).collect_config_data()

        if cmd_config.out_header_file:
            HeaderWriter(cmd_config.out_header_file).write(config)
        if cmd_config.out_json_file:
            JsonWriter(cmd_config.out_json_file).write(config)
        if cmd_config.out_cmake_file:
            CMakeWriter(cmd_config.out_cmake_file).write(config)
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, GenerateCommandConfig)
